@@ -12,7 +12,7 @@ use crate::check::inappropriate_handshake_message;
 use crate::client::common::{ClientAuthDetails, ClientHelloDetails, ServerCertDetails};
 use crate::client::ech::{self, EchState, EchStatus};
 use crate::client::{hs, ClientConfig, ClientSessionStore};
-use crate::common_state::{CommonState, HandshakeKind, Protocol, Side, State};
+use crate::common_state::{CommonState, HandshakeKind, KxState, Protocol, Side, State};
 use crate::conn::ConnectionRandoms;
 use crate::crypto::ActiveKeyExchange;
 use crate::enums::{
@@ -20,7 +20,6 @@ use crate::enums::{
 };
 use crate::error::{Error, InvalidMessage, PeerIncompatible, PeerMisbehaved};
 use crate::hash_hs::{HandshakeHash, HandshakeHashBuffer};
-#[cfg(feature = "logging")]
 use crate::log::{debug, trace, warn};
 use crate::msgs::base::{Payload, PayloadU8};
 use crate::msgs::ccs::ChangeCipherSpecPayload;
@@ -62,7 +61,7 @@ static DISALLOWED_TLS13_EXTS: &[ExtensionType] = &[
 
 pub(super) fn handle_server_hello(
     config: Arc<ClientConfig>,
-    cx: &mut ClientContext,
+    cx: &mut ClientContext<'_>,
     server_hello: &ServerHelloPayload,
     mut resuming_session: Option<persist::Tls13ClientSessionValue>,
     server_name: ServerName<'static>,
@@ -73,7 +72,7 @@ pub(super) fn handle_server_hello(
     mut hello: ClientHelloDetails,
     our_key_share: Box<dyn ActiveKeyExchange>,
     mut sent_tls13_fake_ccs: bool,
-    server_hello_msg: &Message,
+    server_hello_msg: &Message<'_>,
     ech_state: Option<EchState>,
 ) -> hs::NextStateOrError<'static> {
     validate_server_hello(cx.common, server_hello)?;
@@ -147,6 +146,7 @@ pub(super) fn handle_server_hello(
         KeySchedulePreHandshake::new(suite)
     };
 
+    cx.common.kx_state.complete();
     let shared_secret = our_key_share.complete(&their_key_share.payload.0)?;
 
     let mut key_schedule = key_schedule_pre_handshake.into_handshake(shared_secret);
@@ -228,6 +228,7 @@ fn validate_server_hello(
 pub(super) fn initial_key_share(
     config: &ClientConfig,
     server_name: &ServerName<'_>,
+    kx_state: &mut KxState,
 ) -> Result<Box<dyn ActiveKeyExchange>, Error> {
     let group = config
         .resumption
@@ -244,6 +245,7 @@ pub(super) fn initial_key_share(
                 .expect("No kx groups configured")
         });
 
+    *kx_state = KxState::Start(group);
     group.start()
 }
 
@@ -252,7 +254,7 @@ pub(super) fn initial_key_share(
 pub(super) fn fill_in_psk_binder(
     resuming: &persist::Tls13ClientSessionValue,
     transcript: &HandshakeHashBuffer,
-    hmp: &mut HandshakeMessagePayload,
+    hmp: &mut HandshakeMessagePayload<'_>,
 ) -> KeyScheduleEarly {
     // We need to know the hash function of the suite we're trying to resume into.
     let suite = resuming.suite();
@@ -1554,6 +1556,11 @@ impl State<ClientConnectionData> for ExpectTraffic {
         Ok(self)
     }
 
+    fn send_key_update_request(&mut self, common: &mut CommonState) -> Result<(), Error> {
+        self.key_schedule
+            .request_key_update_and_update_encrypter(common)
+    }
+
     fn export_keying_material(
         &self,
         output: &mut [u8],
