@@ -7,6 +7,7 @@ use super::handshake::{ServerDhParams, ServerKeyExchange, ServerKeyExchangeParam
 use crate::enums::{
     CertificateCompressionAlgorithm, CipherSuite, HandshakeType, ProtocolVersion, SignatureScheme,
 };
+use crate::error::InvalidMessage;
 use crate::msgs::base::{Payload, PayloadU16, PayloadU24, PayloadU8};
 use crate::msgs::codec::{put_u16, Codec, Reader};
 use crate::msgs::enums::{
@@ -685,7 +686,7 @@ fn cert_entry_ocsp_response() {
     });
 }
 
-fn test_cert_extension_getter(typ: ExtensionType, getter: fn(&CertificateEntry) -> bool) {
+fn test_cert_extension_getter(typ: ExtensionType, getter: fn(&CertificateEntry<'_>) -> bool) {
     let mut ce = sample_certificate_payload_tls13()
         .entries
         .remove(0);
@@ -872,7 +873,7 @@ fn cannot_decode_huge_certificate() {
     buf[7] = 0x00;
     buf[8] = 0xff;
     buf[9] = 0xfd;
-    HandshakeMessagePayload::read_bytes(&buf).unwrap();
+    HandshakeMessagePayload::read_bytes(&buf[..0x10000 + 7]).unwrap();
 
     // however 64KB + 1 byte does not
     buf[1] = 0x01;
@@ -881,12 +882,15 @@ fn cannot_decode_huge_certificate() {
     buf[4] = 0x01;
     buf[5] = 0x00;
     buf[6] = 0x01;
-    assert!(HandshakeMessagePayload::read_bytes(&buf).is_err());
+    assert_eq!(
+        HandshakeMessagePayload::read_bytes(&buf[..0x10001 + 7]).unwrap_err(),
+        InvalidMessage::CertificatePayloadTooLarge
+    );
 }
 
 #[test]
 fn can_decode_server_hello_from_api_devicecheck_apple_com() {
-    let data = include_bytes!("hello-api.devicecheck.apple.com.bin");
+    let data = include_bytes!("../testdata/hello-api.devicecheck.apple.com.bin");
     let mut r = Reader::init(data);
     let hm = HandshakeMessagePayload::read(&mut r).unwrap();
     println!("msg: {:?}", hm);
