@@ -1,4 +1,3 @@
-use alloc::borrow::ToOwned;
 use alloc::boxed::Box;
 use alloc::string::ToString;
 use alloc::sync::Arc;
@@ -13,13 +12,12 @@ use super::common::ActiveCertifiedKey;
 use super::hs::{self, ServerContext};
 use super::server_conn::{ProducesTickets, ServerConfig, ServerConnectionData};
 use crate::check::inappropriate_message;
-use crate::common_state::{CommonState, Side, State};
+use crate::common_state::{CommonState, HandshakeKind, Side, State};
 use crate::conn::ConnectionRandoms;
 use crate::crypto::ActiveKeyExchange;
 use crate::enums::{AlertDescription, ContentType, HandshakeType, ProtocolVersion};
 use crate::error::{Error, PeerIncompatible, PeerMisbehaved};
 use crate::hash_hs::HandshakeHash;
-#[cfg(feature = "logging")]
 use crate::log::{debug, trace};
 use crate::msgs::base::Payload;
 use crate::msgs::ccs::ChangeCipherSpecPayload;
@@ -38,6 +36,7 @@ mod client_hello {
     use pki_types::CertificateDer;
 
     use super::*;
+    use crate::common_state::KxState;
     use crate::crypto::SupportedKxGroup;
     use crate::enums::SignatureScheme;
     use crate::msgs::enums::{ClientCertificateType, Compression, ECPointFormat};
@@ -64,8 +63,8 @@ mod client_hello {
         pub(in crate::server) fn handle_client_hello(
             mut self,
             cx: &mut ServerContext<'_>,
-            server_key: ActiveCertifiedKey,
-            chm: &Message,
+            server_key: ActiveCertifiedKey<'_>,
+            chm: &Message<'_>,
             client_hello: &ClientHelloPayload,
             selected_kxg: &'static dyn SupportedKxGroup,
             sigschemes_ext: Vec<SignatureScheme>,
@@ -192,6 +191,9 @@ mod client_hello {
                 self.session_id = SessionId::random(self.config.provider.secure_random)?;
             }
 
+            cx.common.kx_state = KxState::Start(selected_kxg);
+            cx.common.handshake_kind = Some(HandshakeKind::Full);
+
             self.send_ticket = emit_server_hello(
                 &self.config,
                 &mut self.transcript,
@@ -290,6 +292,7 @@ mod client_hello {
             cx.common
                 .start_encryption_tls12(&secrets, Side::Server);
             cx.common.peer_certificates = resumedata.client_cert_chain;
+            cx.common.handshake_kind = Some(HandshakeKind::Resumed);
 
             if self.send_ticket {
                 let now = self.config.current_time()?;
@@ -377,13 +380,11 @@ mod client_hello {
     }
 
     fn emit_cert_status(transcript: &mut HandshakeHash, common: &mut CommonState, ocsp: &[u8]) {
-        let st = CertificateStatus::new(ocsp.to_owned());
-
         let c = Message {
             version: ProtocolVersion::TLSv1_2,
             payload: MessagePayload::handshake(HandshakeMessagePayload {
                 typ: HandshakeType::CertificateStatus,
-                payload: HandshakePayload::CertificateStatus(st),
+                payload: HandshakePayload::CertificateStatus(CertificateStatus::new(ocsp)),
             }),
         };
 
@@ -613,6 +614,7 @@ impl State<ServerConnectionData> for ExpectClientKx<'_> {
             self.randoms,
             self.suite,
         )?;
+        cx.common.kx_state.complete();
 
         self.config.key_log.log(
             "CLIENT_RANDOM",
