@@ -15,7 +15,6 @@ use crate::conn::ConnectionRandoms;
 use crate::enums::{AlertDescription, ContentType, HandshakeType, ProtocolVersion};
 use crate::error::{Error, InvalidMessage, PeerIncompatible, PeerMisbehaved};
 use crate::hash_hs::HandshakeHash;
-#[cfg(feature = "logging")]
 use crate::log::{debug, trace, warn};
 use crate::msgs::codec::{Codec, Reader};
 use crate::msgs::enums::KeyUpdateRequest;
@@ -89,7 +88,7 @@ mod client_hello {
         fn check_binder(
             &self,
             suite: &'static Tls13CipherSuite,
-            client_hello: &Message,
+            client_hello: &Message<'_>,
             psk: &[u8],
             binder: &[u8],
         ) -> bool {
@@ -129,8 +128,8 @@ mod client_hello {
         pub(in crate::server) fn handle_client_hello(
             mut self,
             cx: &mut ServerContext<'_>,
-            server_key: ActiveCertifiedKey,
-            chm: &Message,
+            server_key: ActiveCertifiedKey<'_>,
+            chm: &Message<'_>,
             client_hello: &ClientHelloPayload,
             selected_kxg: &'static dyn SupportedKxGroup,
             mut sigschemes_ext: Vec<SignatureScheme>,
@@ -501,6 +500,7 @@ mod client_hello {
         let (share, kxgroup) = share_and_kxgroup;
         debug_assert_eq!(kxgroup.name(), share.group);
         let ckx = kxgroup.start_and_complete(&share.payload.0)?;
+        cx.common.kx_state.complete();
 
         extensions.push(ServerExtension::KeyShare(KeyShareEntry::new(
             ckx.group,
@@ -1491,7 +1491,7 @@ impl ExpectTraffic {
 impl State<ServerConnectionData> for ExpectTraffic {
     fn handle<'m>(
         mut self: Box<Self>,
-        cx: &mut ServerContext,
+        cx: &mut ServerContext<'_>,
         m: Message<'m>,
     ) -> hs::NextStateOrError<'m>
     where
@@ -1536,6 +1536,11 @@ impl State<ServerConnectionData> for ExpectTraffic {
             .extract_secrets(Side::Server)
     }
 
+    fn send_key_update_request(&mut self, common: &mut CommonState) -> Result<(), Error> {
+        self.key_schedule
+            .request_key_update_and_update_encrypter(common)
+    }
+
     fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
         self
     }
